import {
    AbstractFactory,
    Container,
    DependencyRequirement,
    PropertyDependency,
    isInstantiable,
    DEPENDENCY_KEYS_METADATA_KEY,
    DEPENDENCY_KEYS_PROPERTY_METADATA_KEY, Instantiable, FactoryProducer,
} from '../../di'
import {Collection, ErrorWithContext} from '../../util'
import {Config} from '../../service/Config'
import {AuthenticatableRepository} from '../types'
import {ORMUserRepository} from './orm/ORMUserRepository'

/**
 * A dependency injection factory that matches the abstract ClientRepository class
 * and produces an instance of the configured repository driver implementation.
 */
@FactoryProducer()
export class AuthenticatableRepositoryFactory extends AbstractFactory<AuthenticatableRepository> {
    protected get config(): Config {
        return Container.getContainer().make<Config>(Config)
    }

    produce(): AuthenticatableRepository {
        return new (this.getAuthenticatableRepositoryClass())()
    }

    match(something: unknown): boolean {
        return something === AuthenticatableRepository
    }

    getDependencyKeys(): Collection<DependencyRequirement> {
        const meta = Reflect.getMetadata(DEPENDENCY_KEYS_METADATA_KEY, this.getAuthenticatableRepositoryClass())
        if ( meta ) {
            return meta
        }
        return new Collection<DependencyRequirement>()
    }

    getInjectedProperties(): Collection<PropertyDependency> {
        const meta = new Collection<PropertyDependency>()
        let currentToken = this.getAuthenticatableRepositoryClass()

        do {
            const loadedMeta = Reflect.getMetadata(DEPENDENCY_KEYS_PROPERTY_METADATA_KEY, currentToken)
            if ( loadedMeta ) {
                meta.concat(loadedMeta)
            }
            currentToken = Object.getPrototypeOf(currentToken)
        } while (Object.getPrototypeOf(currentToken) !== Function.prototype && Object.getPrototypeOf(currentToken) !== Object.prototype)

        return meta
    }

    /**
     * Return the instantiable class of the configured user repository backend.
     * @protected
     */
    protected getAuthenticatableRepositoryClass(): Instantiable<AuthenticatableRepository> {
        const AuthenticatableRepositoryClass = this.config.get('auth.storage', ORMUserRepository)

        if ( !isInstantiable(AuthenticatableRepositoryClass) || !(AuthenticatableRepositoryClass.prototype instanceof AuthenticatableRepository) ) {
            const e = new ErrorWithContext('Provided client repository class does not extend from @extollo/lib.AuthenticatableRepository')
            e.context = {
                configKey: 'auth.storage',
                class: AuthenticatableRepositoryClass.toString(),
            }
        }

        return AuthenticatableRepositoryClass
    }
}
